# -*- coding: utf-8 -*-
#
#  yaya.py - a (Real) YAYA loader for ninix
#  Copyright (C) 2004 by linjian
#  Copyright (C) 2004-2011 by Shyouzou Sugitani <shy@users.sourceforge.jp>
#  Copyright (C) 2011 by henryhu
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License (version 2) as
#  published by the Free Software Foundation.  It is distributed in the
#  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the GNU General Public License for more details.
#

import os
import sys
import logging
import traceback
from ctypes import cdll, c_char_p, c_long, create_string_buffer, byref
import _ctypes
try:
    _yaya = cdll.LoadLibrary('libaya5.so')
except:
    print 'yaya load fail!'
#    traceback.print_exc()
    _yaya = None


class Shiori:
    def __init__(self, dll_name):
        self.dll_name = dll_name
        self.pathdic = []
        self.reqdic = []

    def use_saori(self, saori):
        self.saori = saori

    def find(self, topdir, dll_name):
        result = 0
        if _yaya:
            if os.path.isfile(os.path.join(topdir, 'yaya.txt')):
                result = 205
            elif dll_name is not None and \
                 os.path.isfile(os.path.join(topdir,
                                             ''.join((dll_name[:-3], 'txt')))):
                result = 105
        return result

    def show_description(self):
        logging.info(
            'Shiori: Real Direct YAYA loader for ninix\n'
            '        Copyright (C) 2004 by linjian\n'
            '        Copyright (C) 2004-2011 by Shyouzou Sugitani\n'
            '        Copyright (C) 2011 by henryhu')

    def get_charset(self, topdir, dll_name):
        charset_output = charset_saori = 'default' # XXX
        with open(os.path.join(topdir, ''.join((dll_name[:-3], 'txt')))) as f:
            for line in f:
                if 'charset' in line:
                    charset = line.split(',')[1]
                    charset = charset.strip()
                    charset_output = charset
                    charset_saori = charset
                    break
            for line in f:
                if 'charset.output' in line:
                    charset = line.split(',')[1]
                    charset = charset.strip()
                    charset_output = charset
                    break
            if charset_output == 'default':
                charset_output = 'UTF-8'
            for line in f:
                if 'charset.extension' in line:
                    charset = line.split(',')[1]
                    charset = charset.strip()
                    charset_saori = charset
                    break
            if charset_saori == 'default':
                charset_saori = 'UTF-8'
        return charset_output, charset_saori

    def load(self, topdir):
        self.dir = topdir
        self.charset, self.charset_saori = self.get_charset(self.dir, self.dll_name)
        if _yaya:
            if self.dir.endswith(os.sep):
                topdir = self.dir
            else:
                topdir = ''.join((self.dir, os.sep))
            path = create_string_buffer(topdir)
            self.pathdic += [path] # so python would not free it
#            print self.pathdic
            # since yaya would free it (...)
            # we must not allow python to free it
            ret = _yaya.load(path, len(topdir))
#            print 'load result: %d' % ret
            ## FIXME: ctypes.CFUNCTYPE
            ##_yaya.setcallback(self.saori_exist,
            ##                  self.saori_load,
            ##                  self.saori_unload,
            ##                  self.saori_request)
            return ret
        else:
            return 0

    def unload(self):
        global _yaya
        if _yaya:
            _yaya.unload()
            # force Python to unload the library
            # else the next load would not function properly
            try:
                _ctypes.FreeLibrary(_yaya._handle)
            except:
                pass
            try:
                _ctypes.dlclose(_yaya._handle)
            except:
                pass
            _yaya = cdll.LoadLibrary('libaya5.so')

    __req_charset = 'GET SHIORI/3.0\r\n' \
                    'ID: charset\r\n' \
                    'Sender: ninix-aya\r\n' \
                    'SecurityLevel: local\r\n\r\n'

    def request(self, req_string):
        if _yaya:
            if req_string == self.__req_charset: # XXX
                return 'SHIORI/3.0 200 OK\r\n' \
                       'Sender: YAYA\r\n' \
                       'Value: %s\r\n\r\n' % self.charset
            reqf = _yaya.request
            reqf.restype = c_char_p
            request = create_string_buffer(req_string)
            # since yaya may free it (...)
            # we must not allow python to free it
            self.reqdic += [request] # so python would not free it
            rlen = c_long(len(request))
#            print 'request:'
#            print req_string
            ret = reqf(request, byref(rlen))
#            print 'result len: %d' % rlen.value
#            print ret
            return ret
        else:
            return '' # FIXME

##    def saori_exist(self, saori):
##        module = self.saori.request(saori)
##        if module:
##            self.saori_list[saori] = [module, 0]
##            return len(self.saori_list)
##        else:
##            return 0
##
##    def saori_load(self, saori, path):
##        result = 0
##        if saori in self.saori_list and self.saori_list[saori][1] == 0:
##            result = self.saori_list[saori][0].load(path)
##            self.saori_list[saori][1] = result
##        return result
##
##    def saori_unload(self, saori):
##        result = 0
##        if saori in self.saori_list and self.saori_list[saori][1] != 0:
##            result = self.saori_list[saori][0].unload()
##            self.saori_list[saori][1] = 0
##        return result
##
##    def saori_request(self, saori, req):
##        result = 'SAORI/1.0 500 Internal Server Error'
##        if saori in self.saori_list:
##            if self.saori_list[saori][1] == 0:
##                head, tail = os.path.split(saori)
##                self.saori_list[saori][1] = \
##                                          self.saori_list[saori][0].load(head)
##            if self.saori_list[saori][1]:
##                result = self.saori_list[saori][0].request(req)
##        return result
